/*

a:


b:


c:


d:


 */

#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <math.h>
#include "wrap/thread.h"
#include "wrap/synch.h"
#include "wrap/atomics.h"
#include "wrap/timer.h"


/**
 * Datastruktur som representerar en plats på tåget.
 *
 * Du kan anta att denna datastruktur bara används från funktioner som
 * börjar med 'seat_' och 'train_'. Dvs. datastrukturen är intern.
 */
struct seat {
    // Array som beskriver vem som har bokat platsen i varje delsträcka
    // av tågets färd. Tal mindre än 0 betyder att platsen är ledig på
    // sträckan i fråga. Ett tal större än eller lika med noll betyder
    // att kunden med det kundnumret har bokat platsen.
    int *sections;

    // Antal delsträckor i 'sections'-arrayen.
    int section_count;
};

// Initiera en plats med ett förutbestämt antal delsträckor.
static void seat_init(struct seat *seat, int sections) {
    seat->sections = malloc(sizeof(int) * sections);
    seat->section_count = sections;

    for (int i = 0; i < sections; i++) {
        seat->sections[i] = -1;
    }
}

// Destruera en plats. Avallokerar *inte* seat i och med att 'init'
// inte allokerar minne.
static void seat_destroy(struct seat *seat) {
    free(seat->sections);
}

// Inspektera ifall en plats är bokad eller inte. Returnerar -1 om
// platsen är obokad på den angivna delsträckan, annars returneras vem
// som har bokat platsen.
static int seat_is_occupied(struct seat *seat, int section) {
    return seat->sections[section];
}

// Försök att boka en plats på en delsträcka. 'customer_id' anger vem
// som ska boka platsen, 'section_from' och 'section_to' anger vilken
// delsträcka som ska bokas. 'section_from' är alltid mindre än
// 'section_to'. Returnerar 'true' om platsen var ledig på den angivna
// delsträckan. Annars returneras 'false'.
static bool seat_book(struct seat *seat, int customer_id, int station_from, int station_to) {
    bool ok = true;
    int filled;

    // Försök boka alla delsegment.
    for (filled = station_from; filled < station_to; filled++) {
        if (seat->sections[filled] < 0) {
            seat->sections[filled] = customer_id;
        } else {
            ok = false;
            break;
        }
    }

    // Om platsen inte var ledig hela sträckan, avboka de delar vi redan har bokat.
    if (!ok) {
        for (int i = station_from; i < filled; i++) {
            seat->sections[i] = -1;
        }
    }

    return ok;
}


/**
 * Datastruktur som beskriver alla platser på ett tåg.
 *
 * Du kan anta att endast funktioner som börjar på 'train_' kan komma
 * åt medlemmarna i datastrukturen. Däremot är alla funktioner
 * 'train_' åtkomliga till alla delar av programmet.
 */
struct train {
    // Alla platser.
    struct seat *seats;

    // Antal platser.
    int seat_count;

    // Antal delsträckor som tåget ska åka.
    int section_count;
};

// Skapa en instans av datastrukturen 'train'. Tåget har 'seats'
// platser, och ska åka 'sections' delsträckor.
struct train *train_create(int seats, int sections) {
    struct train *train = malloc(sizeof(struct train));

    train->seat_count = seats;
    train->section_count = sections;
    train->seats = malloc(sizeof(struct seat) * seats);

    for (int i = 0; i < seats; i++) {
        seat_init(&train->seats[i], sections);
    }

    return train;
}

// Förstör ett tåg.
void train_free(struct train *train) {
    for (int i = 0; i < train->seat_count; i++) {
        seat_destroy(&train->seats[i]);
    }
    free(train->seats);
    free(train);
}

// Boka en plats på en delsträcka. Om en ledig plats fanns på en
// delsträcka bokas platsen, och platsens nummer returneras. Annars
// bokas ingenting, och -1 returneras.
int train_book(struct train *train, int customer_id, int station_from, int station_to) {
    for (int i = 0; i < train->seat_count; i++) {
        if (seat_book(&train->seats[i], customer_id, station_from, station_to)) {
            return i;
        }
    }
    return -1;
}

// Hjälpfunktion för att skriva ut en horisontell linje.
static void print_hline(int length) {
    for (int i = 0; i < length; i++)
        printf("-");
    printf("\n");
}

// Skriv ut en karta över vilka platser som är bokade på vilka
// delsträckor i terminalen. Du kan anta att maximalt en tråd kör
// funktionen samtidigt, men den kan köras samtidigt som en bokning
// görs.
void train_print(struct train *train) {
    print_hline(1 + train->seat_count * 3);

    for (int seg = 0; seg < train->section_count; seg++) {
        printf("|");
        for (int i = 0; i < train->seat_count; i++) {
            int occupied = seat_is_occupied(&train->seats[i], seg);
            if (occupied >= 0)
                printf("%2d|", occupied);
            else
                printf("  |");
        }

        printf("\n");

        print_hline(1 + train->seat_count * 3);
    }
}


/**
 * Huvudprogram.
 *
 * Programmet finns för att visa hur koden ovan kan köras. Den är inte
 * nödvändigtvis designad för att visa alla möjliga problem i
 * implementationen.
 *
 * Ändringar nedanför denna kommentar kommer att ignoreras vid
 * bedömning av din lösning. Uppgiften ska alltså lösas utan att lägga
 * till synkronisering här nedanför. Det går självklart bra att
 * modifiera koden nedanför för att testa din lösning.
 */

struct semaphore done;

void worker(void *data) {
    struct train *train = data;

    train_book(train, 10, 0, 8);
    train_book(train, 11, 4, 10);

    sema_up(&done);
}


int main(void) {
    sema_init(&done, 0);
    struct train *train = train_create(5, 10);
    thread_new(&worker, train);

    train_book(train, 1, 0, 8);
    train_book(train, 2, 0, 4);
    train_book(train, 3, 4, 9);

    sema_down(&done);
    train_print(train);

    train_free(train);

    return 0;
}
